@@ -1,9 +1,12 @@
 from sqlalchemy.testing import fixtures, eq_
 from sqlalchemy.testing import AssertsCompiledSQL, assert_raises_message
-from sqlalchemy.sql import table, column, select, func, literal, exists, and_
+from sqlalchemy.sql import table, column, select, func, literal, exists, \
+    and_, bindparam
 from sqlalchemy.dialects import mssql
 from sqlalchemy.engine import default
 from sqlalchemy.exc import CompileError
+from sqlalchemy.sql.elements import quoted_name
+from sqlalchemy.sql.visitors import cloned_traverse
 
 
 class CTETest(fixtures.TestBase, AssertsCompiledSQL):
@@ -119,6 +122,50 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
             "JOIN parts ON anon_1.part = parts.part "
             "GROUP BY anon_1.sub_part", dialect=mssql.dialect())
 
+    def test_recursive_inner_cte_unioned_to_alias(self):
+        parts = table('parts',
+                      column('part'),
+                      column('sub_part'),
+                      column('quantity'),
+                      )
+
+        included_parts = select([
+            parts.c.sub_part,
+            parts.c.part,
+            parts.c.quantity]).\
+            where(parts.c.part == 'our part').\
+            cte(recursive=True)
+
+        incl_alias = included_parts.alias('incl')
+        parts_alias = parts.alias()
+        included_parts = incl_alias.union(
+            select([
+                parts_alias.c.sub_part,
+                parts_alias.c.part,
+                parts_alias.c.quantity]).
+            where(parts_alias.c.part == incl_alias.c.sub_part)
+        )
+
+        s = select([
+            included_parts.c.sub_part,
+            func.sum(included_parts.c.quantity).label('total_quantity')]).\
+            select_from(included_parts.join(
+                parts, included_parts.c.part == parts.c.part)).\
+            group_by(included_parts.c.sub_part)
+        self.assert_compile(
+            s, "WITH RECURSIVE incl(sub_part, part, quantity) "
+            "AS (SELECT parts.sub_part AS sub_part, parts.part "
+            "AS part, parts.quantity AS quantity FROM parts "
+            "WHERE parts.part = :part_1 UNION "
+            "SELECT parts_1.sub_part AS sub_part, "
+            "parts_1.part AS part, parts_1.quantity "
+            "AS quantity FROM parts AS parts_1, incl "
+            "WHERE parts_1.part = incl.sub_part) "
+            "SELECT incl.sub_part, "
+            "sum(incl.quantity) AS total_quantity FROM incl "
+            "JOIN parts ON incl.part = parts.part "
+            "GROUP BY incl.sub_part")
+
     def test_recursive_union_no_alias_one(self):
         s1 = select([literal(0).label("x")])
         cte = s1.cte(name="cte", recursive=True)
@@ -134,6 +181,21 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
                             "SELECT cte.x FROM cte"
                             )
 
+    def test_recursive_union_alias_one(self):
+        s1 = select([literal(0).label("x")])
+        cte = s1.cte(name="cte", recursive=True)
+        cte = cte.union_all(
+            select([cte.c.x + 1]).where(cte.c.x < 10)
+        ).alias("cr1")
+        s2 = select([cte])
+        self.assert_compile(s2,
+                            "WITH RECURSIVE cte(x) AS "
+                            "(SELECT :param_1 AS x UNION ALL "
+                            "SELECT cte.x + :x_1 AS anon_1 "
+                            "FROM cte WHERE cte.x < :x_2) "
+                            "SELECT cr1.x FROM cte AS cr1"
+                            )
+
     def test_recursive_union_no_alias_two(self):
         """
 
@@ -163,6 +225,26 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
                             "SELECT sum(t.n) AS sum_1 FROM t"
                             )
 
+    def test_recursive_union_alias_two(self):
+        """
+
+        """
+
+        # I know, this is the PG VALUES keyword,
+        # we're cheating here.  also yes we need the SELECT,
+        # sorry PG.
+        t = select([func.values(1).label("n")]).cte("t", recursive=True)
+        t = t.union_all(select([t.c.n + 1]).where(t.c.n < 100)).alias('ta')
+        s = select([func.sum(t.c.n)])
+        self.assert_compile(s,
+                            "WITH RECURSIVE t(n) AS "
+                            "(SELECT values(:values_1) AS n "
+                            "UNION ALL SELECT t.n + :n_1 AS anon_1 "
+                            "FROM t "
+                            "WHERE t.n < :n_2) "
+                            "SELECT sum(ta.n) AS sum_1 FROM t AS ta"
+                            )
+
     def test_recursive_union_no_alias_three(self):
         # like test one, but let's refer to the CTE
         # in a sibling CTE.
@@ -171,7 +253,7 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
         cte = s1.cte(name="cte", recursive=True)
 
         # can't do it here...
-        #bar = select([cte]).cte('bar')
+        # bar = select([cte]).cte('bar')
         cte = cte.union_all(
             select([cte.c.x + 1]).where(cte.c.x < 10)
         )
@@ -187,6 +269,30 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
                             "SELECT cte.x, bar.x FROM cte, bar"
                             )
 
+    def test_recursive_union_alias_three(self):
+        # like test one, but let's refer to the CTE
+        # in a sibling CTE.
+
+        s1 = select([literal(0).label("x")])
+        cte = s1.cte(name="cte", recursive=True)
+
+        # can't do it here...
+        # bar = select([cte]).cte('bar')
+        cte = cte.union_all(
+            select([cte.c.x + 1]).where(cte.c.x < 10)
+        ).alias("cs1")
+        bar = select([cte]).cte('bar').alias("cs2")
+
+        s2 = select([cte, bar])
+        self.assert_compile(s2,
+                            "WITH RECURSIVE cte(x) AS "
+                            "(SELECT :param_1 AS x UNION ALL "
+                            "SELECT cte.x + :x_1 AS anon_1 "
+                            "FROM cte WHERE cte.x < :x_2), "
+                            "bar AS (SELECT cs1.x AS x FROM cte AS cs1) "
+                            "SELECT cs1.x, cs2.x FROM cte AS cs1, bar AS cs2"
+                            )
+
     def test_recursive_union_no_alias_four(self):
         # like test one and three, but let's refer
         # previous version of "cte".  here we test
@@ -234,6 +340,53 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
             "FROM cte WHERE cte.x < :x_2) "
             "SELECT bar.x, cte.x FROM bar, cte")
 
+    def test_recursive_union_alias_four(self):
+        # like test one and three, but let's refer
+        # previous version of "cte".  here we test
+        # how the compiler resolves multiple instances
+        # of "cte".
+
+        s1 = select([literal(0).label("x")])
+        cte = s1.cte(name="cte", recursive=True)
+
+        bar = select([cte]).cte('bar').alias("cs1")
+        cte = cte.union_all(
+            select([cte.c.x + 1]).where(cte.c.x < 10)
+        ).alias("cs2")
+
+        # outer cte rendered first, then bar, which
+        # includes "inner" cte
+        s2 = select([cte, bar])
+        self.assert_compile(s2,
+                            "WITH RECURSIVE cte(x) AS "
+                            "(SELECT :param_1 AS x UNION ALL "
+                            "SELECT cte.x + :x_1 AS anon_1 "
+                            "FROM cte WHERE cte.x < :x_2), "
+                            "bar AS (SELECT cte.x AS x FROM cte) "
+                            "SELECT cs2.x, cs1.x FROM cte AS cs2, bar AS cs1"
+                            )
+
+        # bar rendered, only includes "inner" cte,
+        # "outer" cte isn't present
+        s2 = select([bar])
+        self.assert_compile(s2,
+                            "WITH RECURSIVE cte(x) AS "
+                            "(SELECT :param_1 AS x), "
+                            "bar AS (SELECT cte.x AS x FROM cte) "
+                            "SELECT cs1.x FROM bar AS cs1"
+                            )
+
+        # bar rendered, but then the "outer"
+        # cte is rendered.
+        s2 = select([bar, cte])
+        self.assert_compile(
+            s2, "WITH RECURSIVE bar AS (SELECT cte.x AS x FROM cte), "
+            "cte(x) AS "
+            "(SELECT :param_1 AS x UNION ALL "
+            "SELECT cte.x + :x_1 AS anon_1 "
+            "FROM cte WHERE cte.x < :x_2) "
+            "SELECT cs1.x, cs2.x FROM bar AS cs1, cte AS cs2")
+
     def test_conflicting_names(self):
         """test a flat out name conflict."""
 
@@ -290,6 +443,112 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
                             "FROM regional_sales WHERE "
                             "regional_sales.amount < :amount_2")
 
+    def test_union_cte_aliases(self):
+        orders = table('orders',
+                       column('region'),
+                       column('amount'),
+                       )
+
+        regional_sales = select([
+            orders.c.region,
+            orders.c.amount
+        ]).cte("regional_sales").alias("rs")
+
+        s = select(
+            [regional_sales.c.region]).where(
+            regional_sales.c.amount > 500
+        )
+
+        self.assert_compile(s,
+                            "WITH regional_sales AS "
+                            "(SELECT orders.region AS region, "
+                            "orders.amount AS amount FROM orders) "
+                            "SELECT rs.region "
+                            "FROM regional_sales AS rs WHERE "
+                            "rs.amount > :amount_1")
+
+        s = s.union_all(
+            select([regional_sales.c.region]).
+            where(
+                regional_sales.c.amount < 300
+            )
+        )
+        self.assert_compile(s,
+                            "WITH regional_sales AS "
+                            "(SELECT orders.region AS region, "
+                            "orders.amount AS amount FROM orders) "
+                            "SELECT rs.region FROM regional_sales AS rs "
+                            "WHERE rs.amount > :amount_1 "
+                            "UNION ALL SELECT rs.region "
+                            "FROM regional_sales AS rs WHERE "
+                            "rs.amount < :amount_2")
+
+        cloned = cloned_traverse(s, {}, {})
+        self.assert_compile(cloned,
+                            "WITH regional_sales AS "
+                            "(SELECT orders.region AS region, "
+                            "orders.amount AS amount FROM orders) "
+                            "SELECT rs.region FROM regional_sales AS rs "
+                            "WHERE rs.amount > :amount_1 "
+                            "UNION ALL SELECT rs.region "
+                            "FROM regional_sales AS rs WHERE "
+                            "rs.amount < :amount_2")
+
+    def test_cloned_alias(self):
+        entity = table(
+            'entity', column('id'), column('employer_id'), column('name'))
+        tag = table('tag', column('tag'), column('entity_id'))
+
+        tags = select([
+            tag.c.entity_id,
+            func.array_agg(tag.c.tag).label('tags'),
+        ]).group_by(tag.c.entity_id).cte('unaliased_tags')
+
+        entity_tags = tags.alias(name='entity_tags')
+        employer_tags = tags.alias(name='employer_tags')
+
+        q = (
+            select([entity.c.name])
+            .select_from(
+                entity
+                .outerjoin(entity_tags, tags.c.entity_id == entity.c.id)
+                .outerjoin(employer_tags,
+                           tags.c.entity_id == entity.c.employer_id)
+            )
+            .where(entity_tags.c.tags.op('@>')(bindparam('tags')))
+            .where(employer_tags.c.tags.op('@>')(bindparam('tags')))
+        )
+
+        self.assert_compile(
+            q,
+            'WITH unaliased_tags AS '
+            '(SELECT tag.entity_id AS entity_id, array_agg(tag.tag) AS tags '
+            'FROM tag GROUP BY tag.entity_id)'
+            ' SELECT entity.name '
+            'FROM entity '
+            'LEFT OUTER JOIN unaliased_tags AS entity_tags ON '
+            'unaliased_tags.entity_id = entity.id '
+            'LEFT OUTER JOIN unaliased_tags AS employer_tags ON '
+            'unaliased_tags.entity_id = entity.employer_id '
+            'WHERE (entity_tags.tags @> :tags) AND '
+            '(employer_tags.tags @> :tags)'
+        )
+
+        cloned = q.params(tags=['tag1', 'tag2'])
+        self.assert_compile(
+            cloned,
+            'WITH unaliased_tags AS '
+            '(SELECT tag.entity_id AS entity_id, array_agg(tag.tag) AS tags '
+            'FROM tag GROUP BY tag.entity_id)'
+            ' SELECT entity.name '
+            'FROM entity '
+            'LEFT OUTER JOIN unaliased_tags AS entity_tags ON '
+            'unaliased_tags.entity_id = entity.id '
+            'LEFT OUTER JOIN unaliased_tags AS employer_tags ON '
+            'unaliased_tags.entity_id = entity.employer_id '
+            'WHERE (entity_tags.tags @> :tags) AND '
+            '(employer_tags.tags @> :tags)')
+
     def test_reserved_quote(self):
         orders = table('orders',
                        column('order'),
@@ -319,6 +578,101 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
             '(SELECT "CTE".id AS id FROM "CTE") AS anon_2'
         )
 
+    def test_multi_subq_alias(self):
+        cte = select([literal(1).label("id")]).cte(name='cte1').alias("aa")
+
+        s1 = select([cte.c.id]).alias()
+        s2 = select([cte.c.id]).alias()
+
+        s = select([s1, s2])
+        self.assert_compile(
+            s,
+            "WITH cte1 AS (SELECT :param_1 AS id) "
+            "SELECT anon_1.id, anon_2.id FROM "
+            "(SELECT aa.id AS id FROM cte1 AS aa) AS anon_1, "
+            "(SELECT aa.id AS id FROM cte1 AS aa) AS anon_2"
+        )
+
+    def test_cte_refers_to_aliased_cte_twice(self):
+        # test issue #4204
+        a = table('a', column('id'))
+        b = table('b', column('id'), column('fid'))
+        c = table('c', column('id'), column('fid'))
+
+        cte1 = (
+            select([a.c.id])
+            .cte(name='cte1')
+        )
+
+        aa = cte1.alias('aa')
+
+        cte2 = (
+            select([b.c.id])
+            .select_from(b.join(aa, b.c.fid == aa.c.id))
+            .cte(name='cte2')
+        )
+
+        cte3 = (
+            select([c.c.id])
+            .select_from(c.join(aa, c.c.fid == aa.c.id))
+            .cte(name='cte3')
+        )
+
+        stmt = (
+            select([cte3.c.id, cte2.c.id])
+            .select_from(cte2.join(cte3, cte2.c.id == cte3.c.id))
+        )
+        self.assert_compile(
+            stmt,
+            "WITH cte1 AS (SELECT a.id AS id FROM a), "
+            "cte2 AS (SELECT b.id AS id FROM b "
+            "JOIN cte1 AS aa ON b.fid = aa.id), "
+            "cte3 AS (SELECT c.id AS id FROM c "
+            "JOIN cte1 AS aa ON c.fid = aa.id) "
+            "SELECT cte3.id, cte2.id FROM cte2 JOIN cte3 ON cte2.id = cte3.id"
+        )
+
+    def test_named_alias_no_quote(self):
+        cte = select([literal(1).label("id")]).cte(name='CTE')
+
+        s1 = select([cte.c.id]).alias(name="no_quotes")
+
+        s = select([s1])
+        self.assert_compile(
+            s,
+            'WITH "CTE" AS (SELECT :param_1 AS id) '
+            'SELECT no_quotes.id FROM '
+            '(SELECT "CTE".id AS id FROM "CTE") AS no_quotes'
+        )
+
+    def test_named_alias_quote(self):
+        cte = select([literal(1).label("id")]).cte(name='CTE')
+
+        s1 = select([cte.c.id]).alias(name="Quotes Required")
+
+        s = select([s1])
+        self.assert_compile(
+            s,
+            'WITH "CTE" AS (SELECT :param_1 AS id) '
+            'SELECT "Quotes Required".id FROM '
+            '(SELECT "CTE".id AS id FROM "CTE") AS "Quotes Required"'
+        )
+
+    def test_named_alias_disable_quote(self):
+        cte = select([literal(1).label("id")]).cte(
+            name=quoted_name('CTE', quote=False))
+
+        s1 = select([cte.c.id]).alias(
+            name=quoted_name("DontQuote", quote=False))
+
+        s = select([s1])
+        self.assert_compile(
+            s,
+            'WITH CTE AS (SELECT :param_1 AS id) '
+            'SELECT DontQuote.id FROM '
+            '(SELECT CTE.id AS id FROM CTE) AS DontQuote'
+        )
+
     def test_positional_binds(self):
         orders = table('orders',
                        column('order'),
@@ -480,7 +834,8 @@ class CTETest(fixtures.TestBase, AssertsCompiledSQL):
         self.assert_compile(
             stmt,
             'WITH regional_sales AS (SELECT "order"."order" AS "order" '
-            'FROM "order") oracle suffix  SELECT "order"."order" FROM "order", '
+            'FROM "order") oracle suffix  '
+            'SELECT "order"."order" FROM "order", '
             'regional_sales WHERE "order"."order" > regional_sales."order"',
             dialect='oracle'
         )
